<?php

namespace Drupal\field_updater\Commands;

use Drush\Commands\DrushCommands;
use Drupal\field_updater\FieldUpdaterService;
use Drupal\Core\Entity\EntityTypeManagerInterface;

/**
 * Drush commands for Field Updater.
 */
class FieldUpdaterCommands extends DrushCommands {

  /**
   * The field updater service.
   *
   * @var \Drupal\field_updater\FieldUpdaterService
   */
  protected $updaterService;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs a new FieldUpdaterCommands object.
   */
  public function __construct(FieldUpdaterService $updater_service, EntityTypeManagerInterface $entity_type_manager) {
    $this->updaterService = $updater_service;
    $this->entityTypeManager = $entity_type_manager;
    parent::__construct();
  }

  /**
   * Updates entity fields based on a FieldUpdaterType config entity.
   *
   * @param string $id
   *   The ID of the field_updater_type entity.
   *
   * @command field-updater:update
   * @aliases fu-update
   * @usage drush field-updater:update my_updater_id
   * @description Updates entities based on a FieldUpdaterType config entity in batches.
   */
  public function update($id) {
    /** @var \Drupal\field_updater\Entity\FieldUpdaterType $entity */
    $entity = $this->entityTypeManager->getStorage('field_updater_type')->load($id);
    if (!$entity) {
      $this->logger()->error(dt('FieldUpdaterType entity @id not found.', ['@id' => $id]));
      return;
    }
    $this->updaterService->updateEntities($entity, TRUE);
    $batch =& batch_get();
    if (!empty($batch)) {
      drush_backend_batch_process();
    }
    $this->logger()->success(dt('Initiated batch update for @id.', ['@id' => $id]));
  }

  /**
   * Lists all FieldUpdaterType config entity IDs.
   *
   * @command field-updater:list
   * @aliases fu-list
   * @usage drush field-updater:list
   * @description Displays a list of all FieldUpdaterType config entity IDs.
   */
  public function listUpdaters() {
    $storage = $this->entityTypeManager->getStorage('field_updater_type');
    $ids = $storage->getQuery()
      ->accessCheck(FALSE)
      ->execute();

    if (empty($ids)) {
      $this->logger()->warning(dt('No FieldUpdaterType entities found.'));
      return;
    }

    $entities = $storage->loadMultiple($ids);
    $output = [];

    foreach ($entities as $entity) {
      $output[] = [
        'ID' => $entity->id(),
        'Label' => $entity->label(),
      ];
    }

    // Output as a table.
    $this->output()->writeln(dt('FieldUpdaterType Entities:'));
    $this->io()->table(
      ['ID', 'Label'],
      $output
    );

    $this->logger()->success(dt('Listed @count FieldUpdaterType entities.', ['@count' => count($ids)]));
  }

}
