<?php

namespace Drupal\field_updater\Form;

use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\field_updater\FieldUpdaterTypeInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form for Field updater type entities.
 */
class FieldUpdaterForm extends EntityForm {

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity type bundle info.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $entityTypeBundleInfo;

  /**
   * Constructs a FieldMigrateMappingForm object.
   */
  public function __construct(EntityFieldManagerInterface $entity_field_manager, EntityTypeManagerInterface $entity_type_manager, EntityTypeBundleInfoInterface $entity_type_bundle_info) {
    $this->entityFieldManager = $entity_field_manager;
    $this->entityTypeManager = $entity_type_manager;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_field.manager'),
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);

    /** @var \Drupal\field_updater\Entity\FieldUpdaterType $entity */
    $entity = $this->entity;

    // Label field.
    $form['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Label'),
      '#default_value' => $entity->label(),
      '#required' => TRUE,
    ];

    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $entity->id(),
      '#machine_name' => [
        'exists' => [$this, 'exist'],
        'source' => ['label'],
      ],
      '#disabled' => !$entity->isNew(),
    ];

    // Entity type select field.
    $entity_types = $this->getEntityTypesWithBundles();
    $options = array_map(function ($entity_type) {
      return $entity_type['label'];
    }, $entity_types);

    $form['entity_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Entity type'),
      '#options' => $options,
      '#empty_option' => $this->t('- Select -'),
      '#default_value' => $entity->entity_type,
      '#required' => TRUE,
      '#ajax' => [
        'callback' => [$this, 'updateFields'],
        'wrapper' => 'fields-wrapper',
        'event' => 'change',
      ],
    ];

    // Wrapper for AJAX-updated fields.
    $form['fields_wrapper'] = [
      '#type' => 'container',
      '#attributes' => ['id' => 'fields-wrapper'],
    ];

    // Get entity type from form state or entity.
    $entity_type_id = $form_state->getValue('entity_type') ?: ($entity->entity_type ?? '');
    if (!empty($entity_type_id)) {
      // Bundle select field.
      $bundle_options = $entity_types[$entity_type_id]['bundles'] ?? [];
      $form['fields_wrapper']['bundle'] = [
        '#type' => 'select',
        '#title' => $this->t('Bundle'),
        '#options' => $bundle_options,
        '#empty_option' => $this->t('- Select -'),
        '#default_value' => $entity->bundle ?? '',
        '#required' => TRUE,
        '#ajax' => [
          'callback' => [$this, 'updateFields'],
          'wrapper' => 'fields-wrapper',
          'event' => 'change',
          'progress' => ['type' => 'throbber', 'message' => $this->t('Loading fields...')],
        ],
      ];

      // Get bundle from form state or entity.
      $bundle = $form_state->getValue('bundle') ?: ($entity->bundle ?? '');
      if (!empty($bundle)) {
        // Get field definitions for the selected entity type.
        $field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type_id, $bundle);
        $field_options = [];
        $static_options = [];
        foreach ($field_definitions as $field_name => $definition) {
          if ($definition->getFieldStorageDefinition()->isBaseField() === FALSE) {
            $label = $this->t('@field_label (@field_name)', [
              '@field_label' => $definition->getLabel(),
              '@field_name' => $field_name,
            ]);
            if ($definition->getType() !== 'entity_reference_revisions') {
              $field_options[$field_name] = $label;
            }
            $static_options[$field_name] = $label;
          }
        }
        $form['fields_wrapper']['field_target'] = [
          '#type' => 'select',
          '#title' => $this->t('Target field'),
          '#description' => $this->t('Select the target field to update values to.'),
          '#options' => $field_options,
          '#empty_option' => $this->t('- Select -'),
          '#default_value' => $entity->field_target,
          '#required' => TRUE,
          '#ajax' => [
            'callback' => [$this, 'updateFields'],
            'wrapper' => 'fields-wrapper',
            'event' => 'change',
          ],
          '#states' => [
            'visible' => [
              ':input[name="bundle"]' => ['!value' => ''],
            ],
          ],
        ];
        // Field source select (exclude field_target).
        $field_target = $form_state->getValue('field_target') ?: $entity->field_target;
        $source_options = array_merge($field_options, $static_options);
        if (!empty($field_target)) {
          unset($source_options[$field_target]);
        }
        $form['fields_wrapper']['field_source'] = [
          '#type' => 'select',
          '#title' => $this->t('Source field'),
          '#description' => $this->t('Select a single source field to update values from. To copy values from <em>entity_reference_revision</em> field types, this field is required.'),
          '#options' => $source_options,
          '#empty_option' => $this->t('- Select -'),
          '#default_value' => $entity->field_source,
          '#ajax' => [
            'callback' => [$this, 'updateFields'],
            'wrapper' => 'fields-wrapper',
            'event' => 'change',
          ],
          '#states' => [
            'visible' => [
              ':input[name="bundle"]' => ['!value' => ''],
            ],
          ],
          '#attributes' => [
            'disabled' => empty($source_options),
          ],
        ];
        $field_source = $form_state->getValue('field_source') ?: $entity->field_source;
        $target_type = NULL;
        $target_bundles = [];
        if (!empty($field_source)) {
          $source_definition = $field_definitions[$field_source];
          if ($source_definition && $source_definition->getType() === 'entity_reference_revisions') {
            $target_type = $source_definition->getFieldStorageDefinition()->getSetting('target_type');
            $target_bundles = $source_definition->getSetting('handler_settings')['target_bundles'] ?? [];
          }
        }
        $form['fields_wrapper']['target_type'] = [
          '#type' => 'value',
          '#value' => $target_type,
        ];
        if (empty($target_type)) {
          $form['fields_wrapper']['target_bundles'] = [
            '#type' => 'value',
            '#value' => NULL,
          ];
        }
        else {
          $form['fields_wrapper']['target_bundles'] = [
            '#type' => 'checkboxes',
            '#title' => $this->t('Target @target_type @type', [
              '@target_type' => $target_type,
              '@type' => $this->formatPlural(count($target_bundles), 'type', 'types'),
            ]),
            '#options' => $target_bundles,
            '#default_value' => $entity->target_bundles,
            '#required' => TRUE,
          ];
        }
      }
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state): array {
    $actions = parent::actions($form, $form_state);

    if ($this->entity->isNew()) {
      $actions['submit']['#value'] = $this->t('Save and add mappings');
      $actions['submit']['#submit'][] = '::toMapping';
    }
    else {
      $actions['submit']['#value'] = $this->t('Save updater type');
    }

    return $actions;
  }

  /**
   * Submit for new updaters to redirect to mappings.
   *
   * @param array $form
   *   The form object.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   */
  public function toMapping(array &$form, FormStateInterface $form_state): void {
    $form_state->setRedirectUrl($this->entity->toUrl('mapping'));
  }

  /**
   * AJAX callback to update the fields wrapper.
   */
  public function updateFields(array &$form, FormStateInterface $form_state) {
    return $form['fields_wrapper'];
  }

  /**
   * Checks if a mapping ID already exists.
   */
  public function exist($id) {
    $entity = $this->entityTypeManager->getStorage('field_updater_type')->load($id);
    return (bool) $entity;
  }

  /**
   * Page title callback.
   *
   * @return string
   *   The title of the mapping page.
   */
  public function title(FieldUpdaterTypeInterface $field_updater_type) {
    return $this->t('Edit @label', ['@label' => $field_updater_type->label()]);
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\field_updater\Entity\FieldUpdaterType $entity */
    $entity = $this->entity;
    $original_entity = $this->entityTypeManager->getStorage('field_updater_type')->loadUnchanged($entity->id());
    $values = $form_state->getValues();
    $values['target_bundles'] = !empty($values['target_bundles']) ? array_keys(array_filter($values['target_bundles'])) : [];
    $existing_values = [
      'entity_type' => $original_entity->entity_type,
      'bundle' => $original_entity->bundle,
      'field_target' => $original_entity->field_target,
      'field_source' => $original_entity->field_source,
      'target_type' => $original_entity->target_type,
      'target_bundles' => $original_entity->target_bundles,
    ];
    // Ensure $values only includes the keys we care about to avoid irrelevant
    // differences.
    $submitted_values = array_intersect_key($values, $existing_values);
    $changed = array_diff_assoc($submitted_values, $existing_values);

    $entity->set('entity_type', $values['entity_type']);
    $entity->set('bundle', $values['bundle']);
    $entity->set('field_target', $values['field_target']);
    $entity->set('field_source', $values['field_source'] ?? NULL);
    $entity->set('target_type', $values['target_type'] ?? NULL);
    $entity->set('target_bundles', $values['target_bundles']);
    if (!empty($changed)) {
      $entity->setMappings([]);
    }

    $status = parent::save($form, $form_state);

    $this->messenger()->addStatus(
      $status == SAVED_NEW
        ? $this->t('Created the %label Field updater.', ['%label' => $entity->label()])
        : $this->t('Updated the %label Field updater.', ['%label' => $entity->label()])
    );

    $form_state->setRedirect('entity.field_updater_type.collection');
    return $status;
  }

  /**
   * Helper function to return entity types and bundles.
   *
   * @return array
   *   An array of entity type ids with extra bundle info.
   */
  protected function getEntityTypesWithBundles(): array {
    $types_with_bundles = [];
    foreach ($this->entityTypeManager->getDefinitions() as $entity_name => $entity_definition) {
      if ($entity_definition->entityClassImplements(FieldableEntityInterface::class)) {
        $types_with_bundles[$entity_name] = [
          'machine_name' => $entity_name,
          'label' => $entity_definition->getLabel(),
          'bundles' => [],
        ];
      }
    }
    foreach ($types_with_bundles as $entity_type_key => &$entity_type) {
      foreach ($this->entityTypeBundleInfo->getBundleInfo($entity_type_key) as $bundle => $bundle_info) {
        $field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type_key, $bundle);
        $bundle_fields = [];
        foreach ($field_definitions as $field_name => $definition) {
          if (!$definition->getFieldStorageDefinition()->isBaseField()) {
            $bundle_fields[] = $field_name;
          }
        }
        if (count($bundle_fields) >= 1) {
          $entity_type['bundles'][$bundle] = $bundle_info['label'];
        }
      }
      if (count($entity_type['bundles']) === 0) {
        unset($types_with_bundles[$entity_type_key]);
      }
    }
    return $types_with_bundles;
  }

}
